// Stupid demo. Based on other demos.


#include <ultra64.h>
#include "include/config.h"


enum
{
 VNUT,
 PAUL,
 EXAMPLE1,
 NUM_MODELS
};



//=======================================================================//
//===-[ Global Vars ]-===================================================//
//=======================================================================//

//---=[ Frame buffer stuff ]=---//

#ifdef SCREEN_BPP_8
// 8-bit CFB stuff
u8      *Logical_CFB    = _cfb[0];
u8      *Next_CFB       = _cfb[1];
u8      *Next2_CFB      = _cfb[2];
u16     CFB_Count       = 0 ;      // current logical CFB
u16     CFB_NextCount   = 1 ;      // next logical CFB
u16     CFB_NextCount2  = 2 ;      // second next logical CFB
#else
#ifdef SCREEN_BPP_16
// 16-bit CFB stuff
u16     *Logical_CFB    = _cfb[0];
u16     *Next_CFB       = _cfb[1];
u16     *Next2_CFB      = _cfb[2];
u16     CFB_Count       = 0 ;      // current logical CFB
u16     CFB_NextCount   = 1 ;      // next logical CFB
u16     CFB_NextCount2  = 2 ;      // second next logical CFB
#else
#ifdef SCREEN_BPP_32
// 32-bit CFB stuff
u32     *Logical_CFB    = _cfb[0];
u32     *Next_CFB       = _cfb[1];
u32     *Next2_CFB      = _cfb[2];
u16     CFB_Count       = 0 ;      // current logical CFB
u16     CFB_NextCount   = 1 ;      // next logical CFB
u16     CFB_NextCount2  = 2 ;      // second next logical CFB
#endif
#endif
#endif



//---=[ FIFO size ]=---//

u8*     fifosize;



//---=[ Tasks ]=---//

OSTask  task[NUM_CFB];



//---=[ Dynamic data ]=---//

Dynamic dynamic[NUM_CFB];



//---=[ DLs to display ]=---//

extern Gfx      DL_vnut[];
extern Gfx      DL_paul[];
extern Gfx      DL_example1[];



//---=[ Threads ]=---//

static OSThread mainThread;  
static OSThread gameThread;  



//---=[ Stacks ]=---//

u64 bootStack[STACKSIZE/8];  /* For boot code.  Can reuse on boot code exit. */
u64 mainStack[STACKSIZE/8];  /* Stacks for threads... */
u64 gameStack[STACKSIZE/8];



//---=[ Message queues ]=---//

OSMesg      PiMessages[PI_MSGQUEUE_SIZE];
OSMesgQueue PiMessageQ;
OSMesg      DMAMessages[DMA_MSGQUEUE_SIZE];
OSMesgQueue DMAMessageQ;
OSMesg      RDPDoneMessages[RDPDONE_MSGQUEUE_SIZE];
OSMesgQueue RDPDoneMessageQ;
OSMesg      RSPDoneMessages[RSPDONE_MSGQUEUE_SIZE];
OSMesgQueue RSPDoneMessageQ;
OSMesg      RetraceMessages[RETRACE_MSGQUEUE_SIZE];
OSMesgQueue RetraceMessageQ;
OSMesg      dummyMsg;




//---=[ for senor RSP ]=---//

char     *rsp_static_addr;  /* R4300 addr where rsp_static segment loaded */
char     *rsp_static_end;   /* R4300 addr+1 for end of the load region    */
u32      rsp_static_len ;



//---=[ Viewport structure ]=---//

Vp vp =
{
    SCREEN_WIDTH*2, SCREEN_HEIGHT*2, G_MAXZ/2, 0,  // scale 
    SCREEN_WIDTH*2, SCREEN_HEIGHT*2, G_MAXZ/2, 0,  // translate
};



//---=[ Screen clear color ]=---//

u32     ScreenClearColor = GPACK_RGBA5551(0, 0, 0, 1);






//=========================================================================//
//===-[ Code Start ]-======================================================//
//=========================================================================//




void    boot(void *arg)
{
 void MainProc(void*);
 register int i;

    //=== Init OS ===//
    osInitialize();
  
    //=== Screen off ===//
    osViBlack(TRUE) ;

    //=== Create main thread ===//
    osCreateThread(&mainThread, MAIN_ID, MainProc, arg,
         (void *)(mainStack+STACKSIZE/8), (OSPri) MAIN_PRIORITY);

    //=== Start main thread ===//
    osStartThread(&mainThread);
} 






void    InitGlobalVars()
{
    //=== Init RSP Addresses ===//
    rsp_static_len  = _rsp_staticSegmentRomEnd - _rsp_staticSegmentRomStart;
    rsp_static_addr = (char *) _codeSegmentEnd;
    rsp_static_end  = (char *) _codeSegmentEnd + rsp_static_len;

    fifosize = (u8*)fifobuffer + (sizeof(u64) * FIFOSIZE) + 1;
}





void    MainProc(void *arg)
{
 void GameProc(void*);

    //=== Init some globals ===//
    InitGlobalVars() ;

    //=== Create game thread ===//
    osCreateThread(&gameThread, GAME_ID, GameProc, arg,
         (void *)(gameStack+STACKSIZE/8), GAME_PRIORITY);

    //=== Create video manager ===//
    osCreateViManager(OS_PRIORITY_VIMGR);

    //=== Set video mode ===//
    osViSetMode(&osViModeTable[SCREEN_MODE]);
    osViSetSpecialFeatures(OS_VI_DITHER_FILTER_ON);

    //=== Create PI manager ===//
    osCreatePiManager((OSPri) OS_PRIORITY_PIMGR, &PiMessageQ, PiMessages,
            PI_MSGQUEUE_SIZE);

    //=== Start game thread ===//
    osStartThread(&gameThread);

    //=== Relinquish CPU ===//
    osSetThreadPri(NULL, 0);

    for(;;);
}






void GameProc(void *arg)
{
 void GameLoop();
 OSIoMesg dmaMb;    
 Gfx *l;

    //=== Create DMA message queue ===//
    osCreateMesgQueue(&DMAMessageQ, DMAMessages, DMA_MSGQUEUE_SIZE);

    //=== Create RSP message queue ===//
    osCreateMesgQueue(&RSPDoneMessageQ, RSPDoneMessages,
                RSPDONE_MSGQUEUE_SIZE);
    osSetEventMesg(OS_EVENT_SP, &RSPDoneMessageQ, dummyMsg);

    //=== Create RDP message queue ===//
    osCreateMesgQueue(&RDPDoneMessageQ, RDPDoneMessages,
                RDPDONE_MSGQUEUE_SIZE);
    osSetEventMesg(OS_EVENT_DP, &RDPDoneMessageQ, dummyMsg);

    //=== Create VRT message queue ===//
    osCreateMesgQueue(&RetraceMessageQ, RetraceMessages,
                RETRACE_MSGQUEUE_SIZE);
    osViSetEvent(&RetraceMessageQ, dummyMsg, 1);

    GameLoop();
} 


Dynamic         *generate;      // Dynamic info we're generating now
int             entity=0;       // Current object

void GameLoop()
{
 int            firstframe=1;   // first frame
 Gfx            *gdl_start;     // Start of this frame's dynamic display list */
 Gfx            *gdl_end;       // Current position in dynamic display list
 OSTask         *gentask;       // Task we're generating

 u16            pnorm;          // perpsective norm variable

 float          vnut[3]={0.0, 0.0, 0.0};        // vnut
 float          paul[3]={0.0, 0.0, 0.0};        // paul
 float          example1[3]={0.0, 0.0, 0.0};    // example1

 u8             r=0, g=0, b=0, a=1; // color values

 int            i;              // gen purpose
 u32            model=0, count=0;


    while (1)
    {
     ScreenClearColor = GPACK_RGBA5551(r, g, b, a);
     entity = 0;

     //=== Run task we built during the previous frame ===//
     if(!firstframe)
      osSpTaskStart(gentask);

     //=== Point to new dynamic buffer ===//
     generate = &(dynamic[CFB_NextCount]);

     //=== Set up pointers to GDL ===//
     gdl_start = generate->glist;
     gdl_end = gdl_start;

     //=== Init the GDL ===//
     N64initDL (&gdl_end, gdl_start, rsp_static_addr, generate,
                Next_CFB, YES, YES);

     //=== Set up projection matrix ===//
     guPerspective(&(generate->proj), &pnorm, 45.0, 1.333333,
                        0.1, 500.0, 1.0);
     gSPPerspNormalize(gdl_end++, pnorm);
     guLookAt(&(generate->view),
                0.0, 0.0, 200.0,                // eye
                0.0, 0.0, 0.0,                  // at
                0.0, 1.0, 0.0);                 // up

     gSPMatrix(gdl_end++, OS_K0_TO_PHYSICAL(&(generate->proj)),
                G_MTX_PROJECTION | G_MTX_LOAD | G_MTX_NOPUSH);
     gSPMatrix(gdl_end++, OS_K0_TO_PHYSICAL(&(generate->view)),
                G_MTX_PROJECTION | G_MTX_MUL | G_MTX_NOPUSH);

     //=== Draw models ===//
     if(model == PAUL)
     {
      guPosition( &(generate->xform[entity]),
                paul[0], paul[1], paul[2],               
                0.7,                                        
                0.0, 0.0, 0.0);                             
      gSPMatrix(gdl_end++, OS_K0_TO_PHYSICAL(&(generate->xform[entity])),
               G_MTX_MODELVIEW | G_MTX_LOAD | G_MTX_PUSH);
      gSPDisplayList(gdl_end++, OS_K0_TO_PHYSICAL(DL_paul));
     }
     else if(model == VNUT)
     {
      guPosition( &(generate->xform[entity]),
                vnut[0], vnut[1], vnut[2],               
                0.07,                                        
                0.0, 0.0, 0.0);                             
      gSPMatrix(gdl_end++, OS_K0_TO_PHYSICAL(&(generate->xform[entity])),
               G_MTX_MODELVIEW | G_MTX_LOAD | G_MTX_PUSH);
      gSPDisplayList(gdl_end++, OS_K0_TO_PHYSICAL(DL_vnut));
     }
     else if(model == EXAMPLE1)
     {
      guPosition( &(generate->xform[entity]),
                example1[0], example1[1], example1[2],
                0.07,        
                0.0, 0.0, 0.0);
      gSPMatrix(gdl_end++, OS_K0_TO_PHYSICAL(&(generate->xform[entity])),
               G_MTX_MODELVIEW | G_MTX_LOAD | G_MTX_PUSH);
      gSPDisplayList(gdl_end++, OS_K0_TO_PHYSICAL(DL_example1));
     }
     gSPPopMatrix(gdl_end++, G_MTX_MODELVIEW);
     entity++;

_gdlend:

     //=== End the GDL ===//
     N64terminateDL(&gdl_end);

     //=== Init task ===//
     gentask = &(task[CFB_NextCount]);
     N64initTask(gentask, gdl_end, gdl_start);

     //=== Flush cache ===//
     osWritebackDCacheAll();

     //=== Wait for previous task to finish ===//
     N64waitRSP(firstframe);
     N64waitRDP(firstframe);

_alldone:

     //=== Swap buffers ===//
     N64swapCFB(firstframe);

     //=== Wait for VRT ===//
     N64waitVRT() ;

     firstframe = 0;

     vnut[0]     += 10.0;
     vnut[1]     +=  0.0;
     vnut[2]     +=  0.0;

     paul[0]     += 10.0;
     paul[1]     +=  5.0;
     paul[2]     +=  0.0;

     example1[0] += 2.0;
     example1[1] += 5.0;
     example1[2] += 5.0;

     count++;

     if(count==100)
     {
      model++;
      if(model==NUM_MODELS)
       model=0;
      count=0;
     }
    }

}






